/*
 * Grouper v1.4
 *
 * Grouper--Card Puzzle/Game
 * Copyright  1998, 1999 Daniel McCarty
 *
 * (for revisions, see history.txt)
 * (for legal issues, see grouper.c)
 *
 */


#include "cards.h"


extern PrefType gPrefs;

extern CustomPatternType cardPattern[];
extern IndexedColorType cardColorTable[];

/*
   (thanks to Win95 and PowerPoint97 for some of these patterns)
   Some patterns that didn't make it:
	{0x7831, 0x1387, 0xe1c8, 0x8c1e},	// cargo net
	{0x070e, 0x1c38, 0x70e0, 0xc183},	// diagonals
	{0x0814, 0x2241, 0x8041, 0x2214},	// crosshatch (large)
	{0x44aa, 0x11aa, 0x44aa, 0x11aa},	// crosshatch (small)
	{0x1e8c, 0xd8fd, 0xbf1b, 0x3178},	// daisies
	{0x3e07, 0xe107, 0x3e70, 0xc370},	// dizzy
	{0xfe02, 0xfa8a, 0xbaa2, 0xbe80},	// key
	{0x1010, 0xf101, 0x0101, 0x1f10},	// live wire
	{0xf0f0, 0xf0f0, 0xaa55, 0xaa55},	// plaid
	{0x37b3, 0xdcce, 0x733b, 0xcdec},	// ripples
	{0x80c0, 0x6030, 0x0103, 0x060c},	// steel slots
	{0xae4d, 0xefff, 0x084d, 0xae4d},	// stone
	{0x4582, 0x0100, 0x0182, 0x45aa},	// tiles
	{0xf874, 0x2247, 0x8f17, 0x2271},	// thatches
	{0x4d9a, 0x0855, 0xef9a, 0x4d9a},	// waffle
*/


static void DrawCardMark(MarkType markType, GameColorType markColor, GameColorType cardColor, int x, int y);
static void DrawBitmap(Int idBitmap, SWord x, SWord y);


void DrawCard(CardType card, int x, int y, Boolean fSelected, Boolean fNewDealed) {
// a better way, perhaps, would be to draw cards from DrawGridSpot()
	int i, 
		 markX,
		 markY;
	int markStart[3]    = {MarkSquareX, MarkEllipseX, MarkPeanutX};
	RectangleType frame = {{x, y}, {CardX, CardY}};

	GameColorType cardColor = GetStyleColor(card);
	Word          numMarks  = GetStyleMarks(card);
	GameColorType markColor = GetStyleMarkColor(card);
	MarkType      markType  = GetStyleMarkType(card);

	// draw the card frame
	WinDrawRectangle(&frame, CardDiam);
	// draw or undraw the selection indicator
	if (fSelected)
		WinDrawGrayRectangleFrame(rectangleFrame, &frame);
	else
		WinEraseRectangleFrame(rectangleFrame, &frame);

	// resize the rect to represent the inside of the rect
	// and fill it with the appropriate pattern
	RctInsetRectangle(&frame, 1);
	if (gPrefs.useColor)
	{
		WinSetPattern(cardPattern[GetPattern(colorBlack)]);
		WinSetForeColor(cardColorTable[GetPattern(cardColor)]);
	}
	else
	{
		WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);
		WinSetPattern(cardPattern[GetPattern(cardColor)]);
	}
	WinFillRectangle(&frame, CardDiam);
	WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);

	markX = frame.topLeft.x + markStart[markType];
	markY = frame.topLeft.y;

	// draw all marks on the card
	for (i = 0; i < numMarks; i++) {
		// find equidistant spacing
		markY += (CardY - 2 - numMarks * MarkDY) / (numMarks + 1);
		DrawCardMark(markType, markColor, cardColor, markX, markY);
		// add in width of drawn mark
		markY += MarkDY;
	}

	// mark new dealed cards
	if (fNewDealed)
	{
		RectangleType rt;

		rt.topLeft.x = x + CardX - 2*CardDiam - 2;
		rt.topLeft.y = y + CardY - 2*CardDiam - 2;
		rt.extent.x = 2*CardDiam;
		rt.extent.y = 2*CardDiam;

		WinInvertRectangle(&rt, CardDiam);
	}
}


static
void DrawCardMark(MarkType markType, GameColorType markColor, GameColorType cardColor, int x, int y) {
	RectangleType mark;
	int i;
	int markDiam[4] = {NoDiam, DiamEllipse, NoDiam, NoDiam};

	switch (markType) {
	case markSquare:
		RctSetRectangle(&mark, x, y, MarkDY, MarkDY);
		break;
	case markEllipse:
		RctSetRectangle(&mark, x, y, MarkEllipseDX, MarkDY);
		break;
	case markPeanut:
		if (markType == markPeanut) {
			for (i = 0; i < NumPeanutSquares; i++) {
				DrawCardMark(markPeanutSquare, markColor, cardColor, x, y);

				x += MarkPeanutDX - 1;
				// draw every other peanut square a few pixels lower
				y = y + (i % 2 ? - MarkPeanutDiff : MarkPeanutDiff);
			}
		}
		x -= MarkPeanutDX * (NumPeanutSquares - 1);
		y += MarkPeanutFillerY;
		RctSetRectangle(&mark, x, y, MarkPeanutFillerDX, MarkPeanutFillerDY);
		break;
	case markPeanutSquare:
		RctSetRectangle(&mark, x, y, MarkPeanutDX, MarkPeanutDY);
		break;
	}

	// we don't want to draw the filler bar in markPeanut mode
	if (markType != markPeanut) {
		WinDrawRectangle(&mark, markDiam[markType]);

		// draw white border for black on black or white on black
		// (used to be only for b on b, but w on b just looked too small)
		if (cardColor == colorBlack && markColor != colorGray)
			WinInvertRectangle(&mark, markDiam[markType]);

		// resize the rect to represent the inside of the rect
		// but only if it isn't a black card with a gray mark
		if (cardColor != colorBlack || markColor != colorGray)
			RctInsetRectangle(&mark, 1);
	} else
		RctInsetRectangle(&mark, 1);

	if (gPrefs.useColor)
	{
		WinSetPattern(cardPattern[GetPattern(colorBlack)]);
		WinSetForeColor(cardColorTable[GetPattern(markColor)]);
	}
	else
	{
		WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);
		WinSetPattern(cardPattern[GetPattern(markColor)]);
	}
	WinFillRectangle(&mark, markDiam[markType]);
	WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);
}


void DeckInit(DeckPtr deck) {
	Byte i, j, k, l, m,
		  index;

	CardType tempCard;

	m = 0;
	for (i = 0; i < 3; i++)					// i: card color
		for (j = 0; j < 3; j++)				// j: number of marks
			for (k = 0; k < 3; k++)			// k: mark color
				for (l = 0; l < 3; l++)		// l: mark type
					deck->card[m++] = (i << 6) + (j << 4) + (k << 2) + l;

	// this is a decent way to randomize an array...it's linear too!
	// only shuffle up to numCards; this is an easy way to present 3
	// levels of difficulty without altering the deck
#if EMULATION_LEVEL == EMULATION_NONE
	for (i = 0; i < deck->numCards; i++) {
		index = (UInt)SysRandom(0) % deck->numCards;

		tempCard = deck->card[i];
		deck->card[i] = deck->card[index];
		deck->card[index] = tempCard;
	}
#endif

	SetDeckFullP(deck);
}


void DrawDeck(Boolean fDraw, DeckPtr deck) {
/*
 * This function evaluates whether the deck is full or not.  If it is,
 * it draws the entire deck; if not, it draws the correct number of
 * cards remaining.
 */
	int i;

	RectangleType deckRect = {{DeckX, DeckY}, {DeckDX, DeckDY}};

	if (fDraw) {
		// resize deckRect to represent the base card
		RctSetRectangle(&deckRect, DeckX, DeckCardY, DeckDX, DeckCardDY);

		// draw the main card
		DrawBitmap(idBitmapDeck, deckRect.topLeft.x, deckRect.topLeft.y);

		// draw all remaining card tops
		for (i = 1; i < min(CardsLeftP(deck), DeckMaxCardsDrawn); i++)
			DrawBitmap(idBitmapDeckTop, deckRect.topLeft.x, deckRect.topLeft.y - CardEndY * i);
	} else
		// erase the entire deck
		WinEraseRectangle(&deckRect, NoDiam);
}


void DrawCardsLeft(DeckPtr deck) {
/*
 * This function evaluates whether the deck is full or not.  If it is,
 * it draws the entire deck; if not, it draws the correct number of
 * cards remaining.
 */
	int i;
	char string[3];
	char numericSpaceChr;

	FontID oldFont;
	RectangleType deckRect = {{DeckX, DeckY}, {DeckDX, DeckDY}};

	if (CardsLeftP(deck) < DeckMaxCardsDrawn) {
		// erase the remaining cards
		deckRect.extent.y = CardEndY * (DeckMaxCardsDrawn - CardsLeftP(deck));
		WinEraseRectangle(&deckRect, NoDiam);

		// if there's only 1 card left we don't want to draw over the main card
		if (CardsLeftP(deck) > 1)
			DrawBitmap(idBitmapDeckTop, deckRect.topLeft.x, deckRect.topLeft.y + deckRect.extent.y);
	}

	// draw the cards remaining in inverted bold on the card
	oldFont = FntSetFont(boldFont);

	// erase any possible previous drawing
	ChrNumericSpace(&numericSpaceChr);
	MemSet(string, 2, numericSpaceChr);
	WinDrawInvertedChars(string, 2, DeckNumberX, DeckNumberY);

	StrIToA(string, CardsLeftP(deck));
	// if CardsLeft is less than 10, center the single digit
	i = (CardsLeftP(deck) < 10 ? DeckNumberDX : 0);
	WinDrawInvertedChars(string, StrLen(string), DeckNumberX + i, DeckNumberY);

	FntSetFont(oldFont);
}


static
void DrawBitmap(Int idBitmap, SWord x, SWord y) {
	VoidHand  resH;
	BitmapPtr resP;

	resH = DmGetResource(bitmapRsc, idBitmap);
	if (resH) {
		resP = MemHandleLock(resH);

		WinDrawBitmap(resP, x, y);

		MemPtrUnlock(resP);
		DmReleaseResource(resH);
	}
}


void DrawPatternSample(GameColorType sampleColor, int x, int y) {
	RectangleType spot = {{x, y}, {SampleDX, SampleDY}};

	WinDrawRectangle(&spot, 0);
	RctInsetRectangle(&spot, 1);
	if (gPrefs.useColor)
	{
		WinSetPattern(cardPattern[GetPattern(colorBlack)]);
		WinSetForeColor(cardColorTable[GetPattern(sampleColor)]);
	}
	else
	{
		WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);
		WinSetPattern(cardPattern[GetPattern(sampleColor)]);
	}
	WinFillRectangle(&spot, 0);
	WinSetForeColor(cardColorTable[GetPattern(colorBlack)]);
}

int GetPattern(GameColorType color) {
	switch (color) {
	case colorWhite:
		return PatternWhite;
		break;
	case colorGray:
		return PatternGray + gPrefs.grayPattern;
		break;
	case colorBlack:
		return PatternBlack;
		break;
	}
	
	return PatternBlack;
}

/*
static void DrawDealtCard(int gridSpot);
// for now, don't use this: it looks absolutely hideous; we'd have to
// implement a double-buffered screen for this to look halfway decent,
// and that's a lot of work
static void DrawDealtCard(int gridSpot) {
	// this is a half-width card, halfway between the deck and its
	// dealt location on the grid
	RectangleType halfWay = {
		{(Grid[gridSpot].spot.topLeft.x + gDeckRect.topLeft.x) / 2,
		 (Grid[gridSpot].spot.topLeft.y + gDeckRect.topLeft.y) / 2},
		{CardX / 2, CardY}};

	WinSetClip(&halfWay);
	WinDrawRectangle(&halfWay, CardDiam);
	RectIncrement(&halfWay, 1);
	WinInvertRectangle(&halfWay, CardDiam);
	RectIncrement(&halfWay, 1);
	WinDrawRectangle(&halfWay, CardDiam);
	SysTaskDelay(FrameDelay);
	RectIncrement(&halfWay, -2);
	WinEraseRectangle(&halfWay, CardDiam);
	SysTaskDelay(FrameDelay);
	WinResetClip();
}*/