#ifndef GROUPER_H
#define GROUPER_H

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <PalmChars.h>
#include <DLServer.h>						// for getting user-name


#define AppID			 'dmGR'				// registered with Palm/3Com
#define AppVer			 3

#define NumPicks		 3						// pick per group
#define NumProps		 4						// properties per card
#define NumGrid		 12					// number of spots on the grid

#define NumGameSounds		10
#define MaxNotesPerSound	4

#if EMULATION_LEVEL == EMULATION_NONE
	#define DifficultyFactor 27			// number of cards per level
#else
	#define DifficultyFactor 12			// makes debugging easier
#endif

#if EMULATION_LEVEL == EMULATION_NONE
	#define TextDelay    2
	#define FrameDelay   10					// assuming 10 frames/sec
	#define FramesPerSec 6
#else
	#define TextDelay		1					// since emulation is so slow,
	#define FrameDelay	1					// "remove" these wait times
	#define FramesPerSec	1
#endif	// EMULATION_LEVEL == EMULATION_NONE

// status location definitions
#define StatusX	1
#define StatusY	146
#define StatusDX	108
#define StatusDY	12

// high score location definitions
#define MaxHighScores				5
#define MaxNameLength				9
#define FirstHighScoreY				28
#define HighScoreHeight				12
#define HighScoreNameColumnX		15
#define HighScoreLevelColumnX		84		// right-aligned
#define HighScoreScoreColumnX		123	// right-aligned
#define HighScoreWinColumnX		127
#define HighScoreFlagpoleX			4
#define HighScoreFlagpoleY			19
#define HighScoreFlagpoleDY		54
#define HighScoreFlagpoleDX		2

// scoring definitions
#define ScoreUndo						-7		// -points for an undo
#define ScoreHint						-2		// -points for a hint
#define MaxScoreGroup					15		// max points for a group
#define MinScoreGroup					5		// min points for a group
#define ScoreWin						15		// points for winning the game
#define ScoreX							121
#define ScoreY							1
#define ScoreGroupX						141
#define ScoreGroupY						1

// misc. location/size definitions
#define AboutDeckX				63
#define AboutDeckY				49
#define PrefsSampleX				48
#define PrefsSampleY				72
#define ListPatternSize			8
#define NumAnimationBitmaps	4

// misc. defines
#define DialogButton0		0
#define DialogButton1		1
#define NoCard					-1				// grid spot doesn't have a card
#define MenuHandyman			0				// menu index for pulldown and item
#define UpdateRedrawMask	0x1
#define UpdateNewGameMask	0x2

// this is how we cram 81 styles into 1 byte
// using a bit array is possible (see card.ba and grouper.ba), but isn't
// cost effective at this time; so for now we'll stick with shifts and macros
#define StyleColorMask		0x0003
#define StyleMarksMask		0x000c
#define StyleMarkColorMask	0x0030
#define StyleMarkTypeMask	0x00c0

// deck macros
#define CardsInDeck					81
#define DeckFullIndex				-1
#define DeckFull(deck)				(deck.index == DeckFullIndex)
#define DeckEmpty(deck)				(deck.index == deck.numCards - 1)
#define SetDeckFull(deck)			(deck.index = DeckFullIndex)
#define SetDeckEmpty(deck)			(deck.index = deck.numCards - 1)
#define CardsLeft(deck)				(deck.numCards - deck.index - 1)
// same macros as above, but for pointers--is there a better way?
#define DeckFullP(deckP)			(deckP->index == DeckFullIndex)
#define DeckEmptyP(deckP)			(deckP->index == deckP->numCards - 1)
#define SetDeckFullP(deckP)		(deck->index = DeckFullIndex)
#define CardsLeftP(deckP)			(deckP->numCards - deckP->index - 1)

// card macros
#define GetStyleMarkType(card)	((card & StyleMarkTypeMask) >> 6)
											// 2 hi bits in the style byte
#define GetStyleMarkColor(card)	((card & StyleMarkColorMask) >> 4)
											// next two bits
#define GetStyleMarks(card)		(((card & StyleMarksMask) >> 2) + 1)
#define GetStyleColor(card)		(card & StyleColorMask)
											// 2 lo bits in the style byte


#define min(a, b) ((a < b) ? a : b)
#define max(a, b) ((a > b) ? a : b)

//  .------------.
//  | STRUCTURES |
//  '------------'
typedef enum {
	modeStart = 0,
	modeHint,
	modeButWhy,
	modeEnd
} ModeType;

typedef enum {
	soundShuffle = 0,
	soundDeal,
	soundSelect,
	soundUnselect,
	soundNotAGroup,
	soundGrouper,
	soundGameOver,
	soundYouWin,
	soundNewHighScore,
	soundError
} GameSoundsType;

typedef enum {
	msgClear = 0,								// clear current message
	msgWelcome,									// Welcome to Grouper!
	msgNotAGroup,								// Not a group
	msgYouWin,									// You Win!!!
	msgGameOver,								// No groups: Game Over
	msgGrouper									// Grouper!
} MessageType;

typedef struct {
	Long frequency;							// we don't need an amplitude value
	UInt duration;								// because of the Pilot's weak speaker
} NoteType;

typedef struct {
	UInt numNotes;
	NoteType notes[MaxNotesPerSound];
} SoundType;

typedef struct {
	int grid[NumPicks];
	int count;
} SelectType;

typedef struct {
	Boolean undoAvailable;
	Boolean redrawDeck;
	int grid[NumPicks];
	SByte card[NumPicks];
} UndoType;

typedef struct {
	int index;									// index into grid[] (current hint)
	int grid[NumPicks];						// grid position of hint
	Boolean fUsed[NumPicks];				// flag whether the hint has been used

	int numGroups;								// total number of current groups
} HintType;

typedef struct {
	UInt16 score;
	Byte reasonWhy;							// reason why last group was invalid
	Boolean fSound;							// fSound is stored here because it
													// defaults to the sound pref at start
	ModeType mode;								// game must always be in _some_ mode
	HintType hint;								// current hint info
	SelectType selection;					// current card selection
} GameType;

typedef struct {
	SByte card;

	Boolean fSelected;
	Boolean fNewDealed;
	RectangleType spot;
} GridType;

typedef struct {
	SByte card;
	Boolean fSelected;
	Boolean fNewDealed;
} GridLiteType;								// 3 bytes instead of 12

typedef struct {
	UInt16 score;
	Byte level;
	Boolean fWin;
	char name[MaxNameLength + 1];
} HighScoreType;

typedef Byte CardType;						// unsigned char

typedef struct {
	SByte index;
	Byte numCards;
	CardType card[CardsInDeck];
} DeckType;

typedef DeckType *DeckPtr;

typedef struct {
	Boolean fGameInProgress;
	UInt16 score;
	UndoType undo;
	DeckType deck;
	UInt32 numSecondsSoFar; // number of seconds for the current group, so far
	GridLiteType grid[NumGrid];

	UInt16 lastHighScore;
	Byte difficulty;
	Byte grayPattern;
	
	Boolean useColor;
	Boolean useTimer;

	HighScoreType highScores[MaxHighScores];
} PrefType;

#endif
